@@ -1,7 +1,9 @@
 #!/usr/bin/env python3
-from ._apngasm_python import APNGAsm, APNGFrame, IAPNGAsmListener, create_frame_from_rgb, create_frame_from_rgba
+from __future__ import annotations
+from ._apngasm_python import APNGAsm, APNGFrame, IAPNGAsmListener, create_frame_from_rgb, create_frame_from_rgb_trns, create_frame_from_rgba
 from ._apngasm_python import __version__
 import numpy as np
+import numpy.typing
 from PIL import Image
 from typing import Optional
 
@@ -28,35 +30,35 @@ class APNGAsmBinder:
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.apngasm.reset()
     
-    def frame_pixels_as_pillow(self, frame: int, new_value: Optional[list] = None) -> Optional[Image.Image]:
+    def frame_pixels_as_pillow(self, frame: int, new_value: Optional[Image.Image] = None) -> Optional[Image.Image]:
         '''
         Get/Set the raw pixel data of frame, expressed as a Pillow object.
         This should be set AFTER you set the width, height and color_type.
         
         :param int frame: Target frame number.
-        :param list new_value: (Optional) If set, then the raw pixel data of frame 
+        :param Optional[PIL.Image.Image] new_value: If set, then the raw pixel data of frame 
             is set with this value.
         
         :return: Pillow image object of the frame (get) or None (set)
-        :rtype: PIL.Image.Image or None
+        :rtype: Optional[PIL.Image.Image]
         '''
         if new_value:
-            self.apngasm.get_frames()[frame].pixels = np.array(new_value).flatten()
+            self.apngasm.get_frames()[frame].pixels = np.array(new_value)
         else:
             mode = self.color_type_dict[self.apngasm.get_frames()[frame].color_type]
             return Image.frombytes(mode, (self.apngasm.get_frames()[frame].width, self.apngasm.get_frames()[frame].height), self.apngasm.get_frames()[frame].pixels)
     
-    def frame_pixels_as_numpy(self, frame: int, new_value: Optional[list] = None) -> Optional[np.ndarray]:
+    def frame_pixels_as_numpy(self, frame: int, new_value: Optional[numpy.typing.NDArray] = None) -> Optional[numpy.typing.NDArray]:
         '''
-        Get/Set the raw pixel data of frame, expressed as a 1D numpy array.
+        Get/Set the raw pixel data of frame, expressed as a 3D numpy array.
         This should be set AFTER you set the width, height and color_type.
 
         :param int frame: Target frame number.
-        :param list new_value: (Optional) If set, then the raw pixel data of frame 
+        :param Optional[numpy.typing.NDArray] new_value: If set, then the raw pixel data of frame 
             is set with this value.
 
-        :return: 1D numpy array representation of raw pixel data of frame (get) or None (set)
-        :rtype: numpy.ndarray or None
+        :return: 3D numpy array representation of raw pixel data of frame (get) or None (set)
+        :rtype: Optional[numpy.typing.NDArray]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].pixels = new_value
@@ -68,11 +70,11 @@ class APNGAsmBinder:
         Get/Set the width of frame.
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the width of frame 
+        :param Optional[int] new_value: If set, then the width of frame 
             is set with this value.
 
         :return: width (get) or None (set)
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].width = new_value
@@ -84,11 +86,11 @@ class APNGAsmBinder:
         Get/Set the height of frame.
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the height of frame 
+        :param Optional[int] new_value: If set, then the height of frame 
             is set with this value.
 
         :return: height (get) or None (set)
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].height = new_value
@@ -106,44 +108,45 @@ class APNGAsmBinder:
         6: RGBA (Pillow mode='RGBA')
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the color type of frame 
+        :param Optional[int] new_value: If set, then the color type of frame 
             is set with this value.
         
         :return: color_type of frame (get) or None (set)
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].color_type = new_value
         else:
             return self.apngasm.get_frames()[frame].color_type
     
-    def frame_palette(self, frame: int, new_value: Optional[np.ndarray] = None) -> Optional[np.ndarray]:
+    def frame_palette(self, frame: int, new_value: Optional[numpy.typing.NDArray] = None) -> Optional[numpy.typing.NDArray]:
         '''
         Get/Set the palette data of frame. Only applies to 'P' mode Image (i.e. Not RGB, RGBA)
         Expressed as 2D numpy array in format of [[r0, g0, b0], [r1, g1, b1], ..., [r255, g255, b255]]
         
         :param int frame: Target frame number.
-        :param np.ndarray new_value: (Optional) If set, then the palette data of frame 
+        :param Optional[numpy.typing.NDArray] new_value: If set, then the palette data of frame 
             is set with this value.
 
         :return: 2D numpy array representation of palette data of frame (get) or None (set)
-        :rtype: numpy.ndarray or None
+        :rtype: Optional[numpy.typing.NDArray]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].palette = new_value
         else:
             return self.apngasm.get_frames()[frame].palette
 
-    def frame_transparency(self, frame: int, new_value: Optional[int] = None) -> Optional[np.ndarray]:
+    def frame_transparency(self, frame: int, new_value: Optional[numpy.typing.NDArray] = None) -> Optional[numpy.typing.NDArray]:
         '''
-        Get/Set the transparency data of frame. Expressed as 1D numpy array.
+        Get/Set the color [r, g, b] to be treated as transparent in the frame, expressed as 1D numpy array.
+        For more info, refer to 'tRNS Transparency' in http://www.libpng.org/pub/png/spec/1.2/PNG-Chunks.html
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the transparency of frame 
+        :param Optional[numpy.typing.NDArray] new_value: If set, then the transparency of frame 
             is set with this value.
 
-        :return: 1D numpy array representation of transparency data of frame (get) or None (set)
-        :rtype: numpy.ndarray or None
+        :return: The color [r, g, b] to be treated as transparent in the frame (get) or None (set)
+        :rtype: Optional[numpy.typing.NDArray]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].transparency = new_value
@@ -155,11 +158,11 @@ class APNGAsmBinder:
         Get/Set the palette data size of frame.
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the palette data size of frame 
+        :param Optional[int] new_value: If set, then the palette data size of frame 
             is set with this value.
 
         :return: Palette data size of frame (get) or None (set)
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].palette_size = new_value
@@ -171,11 +174,11 @@ class APNGAsmBinder:
         Get/Set the transparency data size of frame.
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the transparency data size of frame 
+        :param Optional[int] new_value: If set, then the transparency data size of frame 
             is set with this value.
         
         :return: Transparency data size of frame (get) or None (set)
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].transparency_size = new_value
@@ -188,11 +191,11 @@ class APNGAsmBinder:
         Duration of time is delay_num / delay_den seconds.
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the nominator of the duration of frame
+        :param Optional[int] new_value: If set, then the nominator of the duration of frame
             is set with this value.
         
         :return: Nominator of the duration of frame.
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].delay_num = new_value
@@ -205,11 +208,11 @@ class APNGAsmBinder:
         Duration of time is delay_num / delay_den seconds.
         
         :param int frame: Target frame number.
-        :param int new_value: (Optional) If set, then the denominator of the duration of frame
+        :param Optional[int] new_value: If set, then the denominator of the duration of frame
             is set with this value.
         
         :return: Denominator of the duration of frame.
-        :rtype: int or None
+        :rtype: Optional[int]
         '''
         if new_value:
             self.apngasm.get_frames()[frame].delay_den = new_value
@@ -249,7 +252,7 @@ class APNGAsmBinder:
         if pillow_image.mode not in ('RGB', 'RGBA'):
             pillow_image = pillow_image.convert('RGBA')
         return self.add_frame_from_numpy(
-            numpy_data=np.array(pillow_image).flatten(),
+            numpy_data=np.array(pillow_image),
             width=pillow_image.width,
             height=pillow_image.height,
             mode=pillow_image.mode,
@@ -257,19 +260,18 @@ class APNGAsmBinder:
             delay_den=delay_den
         )
     
-    def add_frame_from_numpy(self, numpy_data: int, width: int, height: int, trns_color: int = None, 
-                             mode: str = 'RGBA', delay_num: int = 100, delay_den: int = 1000) -> int:
+    def add_frame_from_numpy(self, numpy_data: numpy.typing.NDArray, width: Optional[int] = None, height: Optional[int] = None, trns_color: Optional[numpy.typing.NDArray] = None, 
+                             mode: Optional[str] = None, delay_num: int = 100, delay_den: int = 1000) -> int:
         '''
         Add frame from numpy array.
         The frame duration is equal to delay_num / delay_den seconds.
         Default frame duration is 100/1000 second, or 0.1 second.
         
-        :param int numpy_data: The pixel data, expressed as 1D numpy array. 
-            For example: [r0, g0, b0, r1, g1, b1, ...]
-        :param int width: The width of the pixel data.
-        :param int height: The height of the pixel data.
-        :param str mode: The color mode of data. Possible values are RGB or RGBA.
-        :param int trns_color: An array of transparency data, expressed as 1D numpy array.
+        :param numpy.typing.NDArray numpy_data: The pixel data, expressed as 3D numpy array.
+        :param Optional[int] width: The width of the pixel data. If not given, the 2nd dimension size of numpy_data is used.
+        :param Optional[int] height: The height of the pixel data. If not given, the 1st dimension size of numpy_data is used.
+        :param Optional[str] mode: The color mode of data. Possible values are RGB or RGBA. If not given, it is determined using the 3rd dimension size of numpy_data.
+        :param Optional[numpy.typing.NDArray] trns_color: The color [r, g, b] to be treated as transparent, expressed as 1D numpy array.
             Only use if RGB mode.
         :param int delay_num: The delay numerator for this frame (defaults to 100).
         :param int delay_den: The delay denominator for this frame (defaults to 1000).
@@ -277,16 +279,40 @@ class APNGAsmBinder:
         :return: The new number of frames.
         :rtype: int
         '''
+        width = width if width else np.shape(numpy_data)[1]
+        height = height if height else np.shape(numpy_data)[0]
+
+        if not mode:
+            if np.shape(numpy_data)[2] == 3:
+                mode = 'RGB'
+            elif np.shape(numpy_data)[2] == 4:
+                mode = 'RGBA'
+            else:
+                raise TypeError('Cannot determine mode from numpy_data. '
+                                'expected 3rd dimension size to be 3 (RGB) or 4 (RGBA). '
+                                f'The given numpy_data 3rd dimension size was {np.shape(numpy_data)[2]}.')
+        
         if mode == 'RGB':
-            frame = create_frame_from_rgb(
-                pixels=numpy_data,
-                width=width,
-                height=height,
-                trns_color=trns_color,
-                delay_num=delay_num,
-                delay_den=delay_den
-            )
+            if type(trns_color) == numpy.typing.NDArray:
+                frame = create_frame_from_rgb_trns(
+                    pixels=numpy_data,
+                    width=width,
+                    height=height,
+                    trns_color=trns_color,
+                    delay_num=delay_num,
+                    delay_den=delay_den
+                )
+            else:
+                frame = create_frame_from_rgb(
+                    pixels=numpy_data,
+                    width=width,
+                    height=height,
+                    delay_num=delay_num,
+                    delay_den=delay_den
+                )
         elif mode == 'RGBA':
+            if type(trns_color) == numpy.typing.NDArray:
+                raise TypeError('Cannot set trns_color on RGBA mode Pillow object. Must be RGB.')
             frame = create_frame_from_rgba(
                 pixels=numpy_data,
                 width=width,
@@ -310,25 +336,25 @@ class APNGAsmBinder:
         '''
         return self.apngasm.assemble(output_path)
     
-    def disassemble_as_numpy(self, file_path: str) -> list:
+    def disassemble_as_numpy(self, file_path: str) -> list[APNGFrame]:
         '''
-        Disassembles an APNG file to a list of frames, expressed as 1D numpy array.
+        Disassembles an APNG file to a list of frames, expressed as 3D numpy array.
         
         :param str file_path: The file path to the PNG image to be disassembled.
         
         :return: A list containing the frames of the disassembled PNG.
-        :rtype: list
+        :rtype: list[apngasm_python._apngasm_python.APNGFrame]
         '''
         return self.apngasm.disassemble(file_path)
 
-    def disassemble_as_pillow(self, file_path: str) -> list:
+    def disassemble_as_pillow(self, file_path: str) -> list[APNGFrame]:
         '''
         Disassembles an APNG file to a list of frames, expressed as Pillow images.
         
         :param str file_path: The file path to the PNG image to be disassembled.
         
         :return: A list containing the frames of the disassembled PNG.
-        :rtype: list
+        :rtype: list[apngasm_python._apngasm_python.APNGFrame]
         '''
         frames_numpy = self.apngasm.disassemble(file_path)
         frames_pillow = []
@@ -350,7 +376,7 @@ class APNGAsmBinder:
         '''
         return self.apngasm.save_pngs(output_dir)
     
-    def load_animation_spec(self, file_path: str) -> list:
+    def load_animation_spec(self, file_path: str) -> list[APNGFrame]:
         '''
         Loads an animation spec from JSON or XML.
         Loaded frames are added to the end of the frame vector.
@@ -358,10 +384,10 @@ class APNGAsmBinder:
         https://github.com/Genshin/PhantomStandards
         You probably won't need to use this function
 
-        :param str file_path: The path of JSON or XML file
+        :param str file_path: The path of JSON or XML file.
 
-        :return: A vector containing the loaded frames
-        :rtype: list
+        :return: A vector containing the loaded frames.
+        :rtype: list[apngasm_python._apngasm_python.APNGFrame]
         '''
         return self.apngasm.load_animation_spec(file_path)
     
@@ -392,12 +418,12 @@ class APNGAsmBinder:
         '''
         return self.apngasm.save_xml(output_path, image_dir)
 
-    def set_apng_asm_listener(self, listener: IAPNGAsmListener = None):
+    def set_apng_asm_listener(self, listener: Optional[IAPNGAsmListener] = None):
         '''
         Sets a listener.
-        You probably won't need to use this function
+        You probably won't need to use this function.
         
-        :param apngasm_python._apngasm_python.IAPNGAsmListener listener: A pointer to the listener object.
+        :param Optional[apngasm_python._apngasm_python.IAPNGAsmListener] listener: A pointer to the listener object.
             If the argument is None,
             a default APNGAsmListener will be created and assigned.
         '''
@@ -419,12 +445,12 @@ class APNGAsmBinder:
         '''
         return self.apngasm.set_skip_first(skip_first)
 
-    def get_frames(self) -> np.ndarray:
+    def get_frames(self) -> list[APNGFrame]:
         '''
         Returns the frame vector.
 
-        :return: frame vector
-        :rtype: numpy.ndarray
+        :return: frame vector.
+        :rtype: list[apngasm_python._apngasm_python.APNGFrame]
         '''
         return self.apngasm.get_frames()
     
@@ -432,7 +458,7 @@ class APNGAsmBinder:
         '''
         Returns the loop count.
 
-        :return: loop count
+        :return: loop count.
         :rtype: int
         '''
         return self.apngasm.get_loops()
@@ -441,7 +467,7 @@ class APNGAsmBinder:
         '''
         Returns the flag of skip first frame.
 
-        :return: flag of skip first frame
+        :return: flag of skip first frame.
         :rtype: int
         '''
         return self.apngasm.get_loops()
@@ -450,7 +476,7 @@ class APNGAsmBinder:
         '''
         Returns the number of frames.
 
-        :return: number of frames
+        :return: number of frames.
         :rtype: int
         '''
         return self.apngasm.frame_count()
@@ -461,7 +487,7 @@ class APNGAsmBinder:
         Leaves the apngasm object in a clean state.
         Returns number of frames disposed of.
 
-        :return: number of frames disposed of
+        :return: number of frames disposed of.
         :rtype: int
         '''
         return self.apngasm.reset()
